#include "header\GBA.h"
#include "header\BG.h"
#include "header\Timer.h"
#include "header\DMA.h"
#include "header\IRQ.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\Blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "data.h"
#include "resource\BitMapFont.h"
#include "resource\tile.map.h"
#include <stdlib.h>
#include <string.h>

//const int __gba_multiboot;  //GBANp

#define MAP_CX 256
#define MAP_CY 256
#define BLOCK_CX 32
#define BLOCK_CY 32

#define MAP_BLOCK_CX  (MAP_CX / BLOCK_CX)
#define MAP_BLOCK_CY  (MAP_CY / BLOCK_CY)
#define MAP_BLOCK_CNT (MAP_BLOCK_CX * MAP_BLOCK_CY)

#define CHR_CORP_FRIEND 1
#define CHR_CORP_ENEMY  2
#define CHR_CORP_SYSTEM 3

#define CHR_FRIEND 2
#define CHR_ENEMY  1
#define CHR_SYSTEM 1
#define CHR_CNT (CHR_FRIEND + CHR_ENEMY + CHR_SYSTEM)

#define GRASS  1	//ړ
#define TOWN   1
#define FOREST 2
#define DESERT 2
#define CASTLE 3
#define MOUNTG 3
#define MOUNTD 3
#define RIVER  99

#define SYSTEM_CHR_CURSER 3
//---------------------------------------------------------------------------
typedef struct {
	u8  CorpFlag;
	u8  VisitFlag;
	u8  Name[12];
	u16 HpVal;
	u8  MapVal;			//}bvW
	u8  MoveVal;		//ړ
	u16 SprNo;			//GBÃXvCgNo
	u16 SprAnime[2];	//Aj[Vp^[
}ST_CHR;

ST_CHR   CHR[CHR_CNT];

//ړ͈́@8*8(256*256 pixsel)
u8 Map[MAP_BLOCK_CNT] = {
   GRASS,  GRASS, GRASS,  RIVER,  RIVER,  RIVER,  GRASS, FOREST,
   GRASS, MOUNTG, GRASS,  RIVER,  RIVER, FOREST,   TOWN,  GRASS,
   GRASS,   TOWN, GRASS,  RIVER,  GRASS,  GRASS,  GRASS,  GRASS,
   GRASS,  GRASS, GRASS,  RIVER,  GRASS,  GRASS,  GRASS,  GRASS,
   GRASS,  GRASS, GRASS, DESERT,  GRASS,  GRASS, FOREST, DESERT,
  MOUNTG,  GRASS, RIVER,  RIVER,  GRASS,  GRASS,  GRASS, DESERT,
  MOUNTG, MOUNTG, RIVER,  RIVER, MOUNTD, DESERT, DESERT, DESERT,
   GRASS,  GRASS, RIVER,  RIVER, MOUNTD, DESERT, DESERT, CASTLE,
};
//LN^[ړ͈́@8*8(256*256 pixsel)
u8 SelectMap[MAP_BLOCK_CNT];

u8 BG2Chr;
u8 BG2Map;
u8 BG3Chr;
u8 BG3Map;
u8 AnimeFlag;
//---------------------------------------------------------------------------
void Init();
void InitBG();
void InitUnit();
void InitSprit();
void InitTimer();

void DispAllUnit();
void DispChrRangeOn();
void DispChrRangeOff();

void CalcChrRange(u8 num);
void CalcChrRangeRe(s8 mapval, s8 moveval);

void InputMoveChr(u8 num);
void MoveChr(u8 num);

void WaitPushA();
void Wait500ms();

void IRQ_USER_Handler();
void IRQ_USER_Setup();
//---------------------------------------------------------------------------
int main(void)
{
	Init();
	InitBG();
	InitUnit();
	InitSprit();
	InitTimer();
	DispAllUnit();

	//BG2 LN^ړ͈͂̕\
	//BG3 n`̕\
	REG_COLEV   = BLEND_BALANCE(0);
	REG_BLDMOD  = BLEND_TOP_BG2 | BLEND_LOW_BG3 | BLEND_MODE_ALPHA;
	REG_DISPCNT = MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D;

	u8 i;
	for(;;){
		for(i=0; i<CHR_FRIEND; i++){
			CalcChrRange(i);
			DispChrRangeOn();

			InputMoveChr(i);
			DispChrRangeOff();

			Wait500ms();
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void Init(){
	Mode3SetBitMapFont((u8*)BitMapFont, 8,8,0);

	BG3Chr = 1;
	BG3Map = 0;
	BG2Chr = 1;
	BG2Map = 2;
	AnimeFlag = 0;
}
//---------------------------------------------------------------------------
void InitBG(){
	BGSetPal256((u16*)EtfsGetDataOffset(_TILE_IMG_PAL_OFFSET));

	//BG3
	REG_BG3CNT = (BG_SIZEA_256_256    | BG_COLOR_256       |
                  BG_CHARBASE(BG3Chr) | BG_MAPBASE(BG3Map) | 0);
	BGSetChrData(BG3Chr, (u16*)EtfsGetDataOffset(_TILE_IMG_BIN_OFFSET), _TILE_IMG_BIN_SIZE);
	BGSetMapData(BG3Map, (u16*)tileData, tileSize*4);

	//BG2
	REG_BG2CNT = (BG_SIZEA_256_256    | BG_COLOR_256       |
                  BG_CHARBASE(BG2Chr) | BG_MAPBASE(BG2Map) | 0);

	u16 *BGMap = MEM_BG_MAP(BG2Map);
	int i;
	for(i=0; i<32*32; i++)	//@MapTile
		*(BGMap+i) = 0x0010;
}
//---------------------------------------------------------------------------
void InitSprit(){
	SpriteInitialize();
	SpriteSetPal256((u16*)EtfsGetDataOffset(_TILE_IMG_PAL_OFFSET));
	SpriteSetData((u16*)EtfsGetDataOffset(_SPRCHR_IMG_BIN_OFFSET), _SPRCHR_IMG_BIN_SIZE);

	int i;
	for(i=0; i<CHR_CNT; i++){
		SpriteSetSize(i, SP_SIZE_32, SP_SQUARE, SP_COLOR_256);
		MoveChr(i);
	}
}
//---------------------------------------------------------------------------
void MoveChr(u8 num){
	u16 x,y;

	x = (CHR[num].MapVal % MAP_BLOCK_CX) * BLOCK_CX;
	y = (CHR[num].MapVal / MAP_BLOCK_CX) * BLOCK_CY;
	SpriteMove(CHR[num].SprNo, x, y);
}
//---------------------------------------------------------------------------
void InitTimer(){
	IRQ_USER_Setup();

	REG_TM0D = 0x8000;
	REG_TM0CNT = TM_ENABLE | TM_FREQ_PER_256 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void InitUnit(){
	strcpy(CHR[0].Name, "@g");
	CHR[0].CorpFlag    = CHR_CORP_FRIEND;
	CHR[0].VisitFlag   = 1;
	CHR[0].HpVal       = 3;
	CHR[0].MapVal      = 8;
	CHR[0].MoveVal     = 4;
	CHR[0].SprNo       = 0;
	CHR[0].SprAnime[0] = 0x00;
	CHR[0].SprAnime[1] = 0x81;

	strcpy(CHR[1].Name, "m");
	CHR[1].CorpFlag    = CHR_CORP_FRIEND;
	CHR[1].VisitFlag   = 1;
	CHR[1].HpVal       = 1;
	CHR[1].MapVal      = 1;
	CHR[1].MoveVal     = 4;
	CHR[1].SprNo       = 1;
	CHR[1].SprAnime[0] = 0x08;
	CHR[1].SprAnime[1] = 0x89;

	strcpy(CHR[2].Name, "G`");
	CHR[2].CorpFlag    = CHR_CORP_ENEMY;
	CHR[2].VisitFlag   = 1;
	CHR[2].HpVal       = 1;
	CHR[2].MapVal      = 30;
	CHR[2].MoveVal     = 0;
	CHR[2].SprNo       = 2;
	CHR[2].SprAnime[0] = 0x101;
	CHR[2].SprAnime[1] = 0x181;

	strcpy(CHR[3].Name, "Cursor");
	CHR[3].CorpFlag    = CHR_CORP_SYSTEM;
	CHR[3].VisitFlag   = 0;
	CHR[3].HpVal       = 0;
	CHR[3].MapVal      = 0;
	CHR[3].MoveVal     = 0;
	CHR[3].SprNo       = 3;
	CHR[3].SprAnime[0] = 0x109;
	CHR[3].SprAnime[1] = 0x189;

	//GB 0x10 0x91
	//GC 0x18 0x99
}
//---------------------------------------------------------------------------
void DispAllUnit(){
	int i;
	for(i=0; i<CHR_CNT; i++){
		if( CHR[i].VisitFlag )
			SpriteSetChr(CHR[i].SprNo, CHR[i].SprAnime[AnimeFlag]);
		else
			SpriteSetChr(CHR[i].SprNo, 0x110); //
	}
}
//---------------------------------------------------------------------------
void DispChrRangeOn(){
	u16 *BGMap = MEM_BG_MAP(BG2Map);
	u16 *BGTmp;
	int i,j,k;

	//LN^̈ړׂ͈͂čFɂ
	for(i=0; i<1024; i++)
		*(BGMap+i) = 0x0010;

	//ړ\͈͂𓧖Fɂ
	for(i=0; i<MAP_BLOCK_CNT; i++){
		if( SelectMap[i] ){
			BGTmp  = BGMap;
			BGTmp += (i%MAP_BLOCK_CX) * 0x04;	//x
			BGTmp += (i/MAP_BLOCK_CX) * 0x80;	//y

			for(j=0; j<4; j++){
				for(k=0; k<4; k++){
					*(BGTmp+k) = 0x0090;
				}
				BGTmp += 0x20;
			}

		}
	}
	//\
	WaitForVsync();
	REG_COLEV = BLEND_BALANCE(8);
}
//---------------------------------------------------------------------------
void DispChrRangeOff(){
	WaitForVsync();
	REG_COLEV = BLEND_BALANCE(0);
}
//---------------------------------------------------------------------------
void CalcChrRange(u8 num){
	int i;

	//@ړ͈̓tO
	for(i=0; i<MAP_BLOCK_CNT; i++){
		SelectMap[i] = 0;
	}

	s8 mapval  = CHR[num].MapVal;
	s8 moveval = CHR[num].MoveVal + Map[mapval];
	CalcChrRangeRe(mapval, moveval);
/*
	for(i=0; i<MAP_BLOCK_CNT; i++){
		dprintf("%2d,",SelectMap[i]);
		if((i+1) % MAP_BLOCK_CX == 0)
			dprintf("\n");
	}
	dprintf("\n");
*/
}
//---------------------------------------------------------------------------
void CalcChrRangeRe(s8 mapval, s8 moveval){

	if(mapval<0 || mapval>MAP_BLOCK_CNT)
		return;

	moveval -= Map[mapval];
	if(moveval<=0)
		return;

	if(SelectMap[mapval] < moveval)
		SelectMap[mapval] = moveval;

	//mapf[^1z񂾂񂾂 傤Ȃ(^^;
	if((mapval+1) % MAP_BLOCK_CX != 0)
		CalcChrRangeRe(mapval+1, moveval);

	if((mapval  ) % MAP_BLOCK_CX != 0)
		CalcChrRangeRe(mapval-1, moveval);

	CalcChrRangeRe(mapval+MAP_BLOCK_CX, moveval);
	CalcChrRangeRe(mapval-MAP_BLOCK_CX, moveval);
}
//---------------------------------------------------------------------------
void IRQ_USER_Handler(){
	REG_IME  = IRQ_MASTER_OFF;
	u16 Flag = REG_IF;

	if( Flag & IRQ_BIT_TIMER0 ){
		AnimeFlag = (AnimeFlag+1) % 2;
		DispAllUnit();

		REG_TM0D = 0x8000;
	}

	REG_IF  = Flag;
	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void InputMoveChr(u8 num){
	u32 KeyPrev = 0;
	u32 KeyNew  = 0;
    u32 KeyLate = 0;

	CHR[SYSTEM_CHR_CURSER].MapVal    = CHR[num].MapVal;
	CHR[SYSTEM_CHR_CURSER].VisitFlag = 1;
	MoveChr(SYSTEM_CHR_CURSER);
	DispAllUnit();

	for(;;){

		//O񉟂ĂȂǁA񉟂ꂽ̂擾
		WaitForVsync();
		MoveChr(SYSTEM_CHR_CURSER);

		KeyPrev = KeyNew;
		KeyNew  = *KEYS  ^ 0x3ff;
		KeyLate = KeyNew ^ (KeyPrev & KeyNew);

		//G[`FbNȂ܂m(__)m
		if( KeyLate & KEY_A ){
			if(SelectMap[ CHR[SYSTEM_CHR_CURSER].MapVal ])
				break;
		}
		if( KeyLate & KEY_UP   ){
			CHR[SYSTEM_CHR_CURSER].MapVal -= MAP_BLOCK_CX;
		}
		if( KeyLate & KEY_DOWN ){
			CHR[SYSTEM_CHR_CURSER].MapVal += MAP_BLOCK_CX;
		}
		if( KeyLate & KEY_LEFT ){
			CHR[SYSTEM_CHR_CURSER].MapVal -= 1;
		}
		if( KeyLate & KEY_RIGHT){
			CHR[SYSTEM_CHR_CURSER].MapVal += 1;
		}

	}//for(;;)

	WaitPushA();

	CHR[SYSTEM_CHR_CURSER].VisitFlag = 0;
	CHR[num].MapVal = CHR[SYSTEM_CHR_CURSER].MapVal;
	MoveChr(num);
	DispAllUnit();
}

//========================================================================
void IRQ_USER_Setup(){
	REG_IME = IRQ_MASTER_OFF;

	REG_INTERUPT = (u32)IRQ_USER_Handler;
	REG_IE |= IRQ_BIT_TIMER0;

	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void Wait500ms(){
	REG_TM1D = 0;
	REG_TM1CNT = TM_FREQ_PER_256 | TM_ENABLE;

	while(REG_TM1D < 0x8000){
		Mode3DrawBitMapStr(1,1,""); //gcc̍œK̂߂ɏȂ݂
									//Ȃ̂ŁAeLg[ȏƂ
	}

	REG_TM1CNT = TM_DISABLE;
	REG_TM1D = 0;
}
